#include "stdafx.h"
#include "KDTree.h"
#include "GraphicObject.h"
#include <algorithm>
#include <math.h>

#pragma warning( disable : 4786)

KDTree::KDTree (DblRect _world, int estNumObj)
  : worldRect(_world)
{
	int threashold = 32;
  root = new KDNode( _world, threashold, true);
}


KDTree::~KDTree()
{
  if (root != 0)
    delete root;   // This should recursvly delete everything
}


void KDTree::InsertObject (GraphicObject* obj, DblRect rect)
{
	root->Insert(obj,rect);
}

void KDTree::RemoveObject (GraphicObject* obj)
{
	root->Remove(obj,obj->GetBoundingRect());
}

list<GraphicObject*> * KDTree::GetObjects (DblRect rect)
{
	list<GraphicObject*> *pList = root->GetObjects(rect);

	if (pList == 0)
	{
		return new list<GraphicObject*>;
	}
	else
	{
		//pList->sort();
		//pList->unique();
		return pList;   
	}
}


KDNode::KDNode(DblRect _space, int _threashold, bool _vert)
  : left(0),
	vertSplit(_vert),
	right(0),
	space(_space),
	threashold(_threashold)

{
	region = new KDRegion();
}


KDNode::~KDNode()
{
	if (left == 0)
		delete left;

	if (right == 0)
		delete right;

	if (region == 0)
		delete region;

	left = right = 0;
	region = 0;
}


void KDNode::Insert(GraphicObject *obj, DblRect rect)
{
	// First determine if this object even belongs in this node
	if (!(rect.Intersect(space)))
	{
		// just return
		return;
	}

	// If we have a region insert into it
	if (region)
	{
		region->AddObject(obj);

		// Determine if we need to split
		if (region->Size() > threashold)
		{
			Split();
		}

	}
	else
	{
		// Insert into left and right
		left->Insert(obj,rect);
		right->Insert(obj,rect);
	}
	
}

void KDNode::Remove(GraphicObject *obj, DblRect rect)
{
	// If this object is not in this area -- return
	if (!(rect.Intersect(space)))
	{
		// just return
		return;
	}

	// If we have a region - simply remove from it
	if (region)
	{
		region->RemoveObject(obj);
	}
	else
	{
		// Otherwise - remove it from left and right
		left->Remove(obj,rect);
		right->Remove(obj,rect);

		// We could merge the regions here -- but we won't for now		
	}
}

// This funciton determines the most efficient split 
// (horizontal or vertical) and divides the node.
void KDNode::Split()
{

	// Lets get our objects	
	list<GraphicObject*> allObjects = region->GetObjects();

	if (vertSplit)
	{
		double vert = space.top + space.Height() / 2;
		left = new KDNode(DblRect(space.left, space.top, space.right, vert), threashold, !vertSplit);
		right = new KDNode(DblRect(space.left, vert, space.right, space.bottom), threashold, !vertSplit);
	}
	else
	{
		double horz = space.left + space.Width() / 2;
		left = new KDNode(DblRect(space.left, space.top, horz, space.bottom), threashold, !vertSplit);
		right = new KDNode(DblRect(horz, space.top, space.right, space.bottom), threashold, !vertSplit);
	}

	// Insert
	for (std::list<GraphicObject*>::iterator iObj = allObjects.begin(); iObj != allObjects.end(); iObj++)
	{
		left->Insert((*iObj), (*iObj)->GetBoundingRect());
		right->Insert((*iObj), (*iObj)->GetBoundingRect());
	}

	// Delete our region
	delete region;
	region = 0;
}

// This is responsible for getting all the objects that intersect with
// the given rectangle.
list<GraphicObject*> * KDNode::GetObjects (DblRect rect)
{
// If this object is not in this area 
	if (!(rect.Intersect(space)))
	{
		// just return an empty list
		return 0;
	}

	// If we have a region - return its list
	if (region)
	{
		return new list<GraphicObject*>(region->GetObjects());
	}
	
	// Otherwise get the left and right - and merge them into one list
	list<GraphicObject*> *pLeftList = left->GetObjects(rect);
	list<GraphicObject*> *pRightList = right->GetObjects(rect);

	// If either are null -- return the other one
	if (pLeftList == 0)
	{
		return pRightList;
	}
	else if (pRightList == 0)
	{
		return pLeftList;
	}

	// Splice, sort, and remove duplicates

   pLeftList->splice(pLeftList->begin(), *pRightList);
   delete pRightList;   
   return pLeftList;
	// Otherwise merge them
	/*std::list<GraphicObject*>::iterator iLeft = pLeftList->begin();
	std::list<GraphicObject*>::iterator iRight = pRightList->begin();

	list <GraphicObject*> *pReturnList = new list<GraphicObject*>;

	while ( (iLeft != pLeftList->end()) && (iRight != pRightList->end()) )
	{		
		if ((*iLeft) < (*iRight))
		{
			pReturnList->push_back(*iLeft);
			iLeft++;
		}
		else if ((*iLeft) > (*iRight))
		{
			pReturnList->push_back(*iRight);
			iRight++;
		}
		else // Must be equal
		{
			// Insert and inc both
			pReturnList->push_back(*iLeft);
			iLeft++;
			iRight++;
		}
	}

	// Add the remaining objects
	while (iLeft != pLeftList->end())
	{		
		pReturnList->push_back(*iLeft);
		iLeft++;	
	}

	while (iRight != pRightList->end())
	{		
		pReturnList->push_back(*iRight);
		iRight++;	
	}

	// Cleanup
	delete pLeftList;
	delete pRightList;

	return pReturnList;*/
}


KDRegion::KDRegion()
{
}


KDRegion::~KDRegion()
{
}


void KDRegion::AddObject (GraphicObject* obj)
{
	objects.insert( pair<GraphicObject*,bool>(obj,true));
}

void KDRegion::RemoveObject (GraphicObject* obj)
{  
	std::map<GraphicObject*,bool>::iterator iObj = objects.find(obj);

	if (iObj != objects.end())
	{
		objects.erase(iObj);
	}
}

list<GraphicObject*> KDRegion::GetObjects ()
{
	list<GraphicObject*> returnList;

	for (std::map<GraphicObject*,bool>::iterator iObj = objects.begin(); iObj != objects.end(); iObj++)
	{
		returnList.push_back( (*iObj).first);
	}

	return returnList;
}