
#include "stdafx.h"
#include "QuadTree.h"
#include "GraphicObject.h"

#include <algorithm>

QuadTree::QuadTree (DblRect _world, int _levels)
  : worldRect(_world), levels(7)
{
  root = new QTNode();
}


QuadTree::~QuadTree()
{
  Clear();

  if (root != 0)
    delete root;   // This should recursvly delete everything
}


void QuadTree::InsertObject (GraphicObject* obj, DblRect rect)
{

   // Make a list of regions that the object should be in and 
   // then add object to each.

   list<QTRegion*> *regionList = new list<QTRegion*>;

   RegionsFromRect(root, worldRect, rect, regionList, true); 
   
   // Go through each affected Region and add this object
   for (std::list<QTRegion*>::iterator i = regionList->begin(); i != regionList->end(); i++)
   {
      (*i)->AddObject(obj);
   }

   // Free up regionlist
   delete regionList;
}

void QuadTree::RemoveObject (GraphicObject* obj)
{
   // Make a list of regions that the object should be in and 
   // then remove object from each.

   std::list<QTRegion*> *regionList = new std::list<QTRegion*>;

   RegionsFromRect(root, worldRect, obj->GetBoundingRect(), regionList); 

   // Go through each affected Region and remove objects
   for (std::list<QTRegion*>::iterator i = regionList->begin(); i != regionList->end(); i++)
   {
      (*i)->RemoveObject(obj);
   }

   // Free up regionlist
   delete regionList;
}

list<GraphicObject*> * QuadTree::GetObjects (DblRect rect)
{
   std::list<GraphicObject*> *returnList;
   returnList = new std::list<GraphicObject*>;

   std::list<QTRegion*> *regionList = new std::list<QTRegion*>;
   RegionsFromRect(root, worldRect, rect, regionList); 

   for (std::list<QTRegion*>::iterator i = regionList->begin(); i != regionList->end(); i++)
   {
	   // Inserts splices returned list with results list
       returnList->splice(returnList->begin(), (*i)->GetObjects());
   }

   delete regionList;

   // Make the return List contain only unique pointers
   returnList->sort();
   returnList->unique();
   return returnList;
}

void QuadTree::Clear ()
{
  // Make a list of all objects then remove.  
   std::list<QTRegion*> *regionList = new std::list<QTRegion*>;
   RegionsFromRect(root, worldRect, worldRect, regionList); 

   // Go through each affected Region and remove objects
   for (std::list<QTRegion*>::iterator i = regionList->begin(); i != regionList->end(); i++)
   {
      (*i)->Clear();
   }

   delete regionList;
}

void QuadTree::RegionsFromRect (QTNode* node, DblRect searchRect, DblRect targetRect, list<QTRegion*>* resultList, bool createNew, int curLevel)
{
  if (curLevel < levels)
  {
	  CPoint  MidPoint = searchRect.CenterPoint(); // Center point of search rect
      int       RegionMask = MASK_ALL_REG;;          // Mask for new region

 
	  // Mask out regions that are not necessary
      if (targetRect.left >= MidPoint.x)
         RegionMask &= MASK_NOT_REG_03;
      else if (targetRect.right <= MidPoint.x)
         RegionMask &= MASK_NOT_REG_12;

      if (targetRect.bottom <= MidPoint.y)      
         RegionMask &= MASK_NOT_REG_23;      
      else if (targetRect.top >= MidPoint.y)      
         RegionMask &= MASK_NOT_REG_01;      


	  // Now recursivly call regions if necessary

	  if (RegionMask & MASK_REG_0) // in region 0
      {
		 DblRect newSearchRect(searchRect.left, searchRect.top, MidPoint.x, MidPoint.y);
         
		 // Add if necessary
         if (node->GetSubNode(QTNode::UPPER_LEFT ) == 0 && createNew == true)         
            node->AddNode(QTNode::UPPER_LEFT, new QTNode);         

		 // As long as there is something there, search the UPPER_LEFT
         if (node->GetSubNode(QTNode::UPPER_LEFT) != 0)
         {
            RegionsFromRect(node->GetSubNode(QTNode::UPPER_LEFT),  // Upper Left Node
				            newSearchRect,                         // Upper Left Rect
				            targetRect,                            // Rectangle searching for
                            resultList,                            // Pointer to results list
							createNew,                             // Create New?
							curLevel+1);                           // Next Level
         }
      }

	  if (RegionMask & MASK_REG_1) // in region 1
      {
		 DblRect newSearchRect(MidPoint.x, searchRect.top, searchRect.right, MidPoint.y);
         
		 // Add if necessary
         if (node->GetSubNode(QTNode::UPPER_RIGHT ) == 0 && createNew == true)         
            node->AddNode(QTNode::UPPER_RIGHT, new QTNode);         

		 // As long as there is something there, search the UPPER_RIGHT
         if (node->GetSubNode(QTNode::UPPER_RIGHT) != 0)
         {
            RegionsFromRect(node->GetSubNode(QTNode::UPPER_RIGHT), // Upper Right Node
				            newSearchRect,                         // Upper Right Rect
				            targetRect,                            // Rectangle searching for
                            resultList,                            // Pointer to results list
							createNew,                             // Create New?
							curLevel+1);                           // Next Level
         }
      }

	  if (RegionMask & MASK_REG_2) // in region 2
      {
		 DblRect newSearchRect(searchRect.left, MidPoint.y, MidPoint.x, searchRect.bottom);
         
		 // Add if necessary
         if (node->GetSubNode(QTNode::LOWER_LEFT ) == 0 && createNew == true)         
            node->AddNode(QTNode::LOWER_LEFT, new QTNode);         

		 // As long as there is something there, search the LOWER_LEFT
         if (node->GetSubNode(QTNode::LOWER_LEFT) != 0)
         {
            RegionsFromRect(node->GetSubNode(QTNode::LOWER_LEFT),  // Lower Left Node
				            newSearchRect,                         // Lower Left Rect
				            targetRect,                            // Rectangle searching for
                            resultList,                            // Pointer to results list
							createNew,                             // Create New?
							curLevel+1);                           // Next Level
         }
      }

	  if (RegionMask & MASK_REG_3) // in region 3
      {
		 DblRect newSearchRect(MidPoint.x, MidPoint.y, searchRect.right, searchRect.bottom);
         
		 // Add if necessary
         if (node->GetSubNode(QTNode::LOWER_RIGHT ) == 0 && createNew == true)         
            node->AddNode(QTNode::LOWER_RIGHT, new QTNode);         

		 // As long as there is something there, search the LOWER_RIGHT
         if (node->GetSubNode(QTNode::LOWER_RIGHT) != 0)
         {
            RegionsFromRect(node->GetSubNode(QTNode::LOWER_RIGHT), // Lower Right Node
				            newSearchRect,                         // Lower Right Rect
				            targetRect,                            // Rectangle searching for
                            resultList,                            // Pointer to results list
							createNew,                             // Create New?
							curLevel+1);                           // Next Level
         }
      }
  }
  else
  {
	  // OK Reached the end!

	  // Add region if necessary
	  if (node->GetRegion() == 0 && createNew == true)
         node->AddRegion();

	  // If region is not nill then add to list (if one exists)
      if (node->GetRegion() != 0 && resultList != 0)      
         resultList->push_back(node->GetRegion());      
  }
}



QTNode::QTNode()
  : region(0)
{
  // Reset all subnodes
  for (int i=0; i <=3; i++)
	  subNodes[i] = 0;
}


QTNode::~QTNode()
{
   for (int i=0; i <=3; i++)
	  if (subNodes[i] != 0)
		  delete subNodes[i];

  if (region != 0)
	  delete region;

}

void QTNode::AddRegion ()
{
  if (region == 0)
    region = new QTRegion;
}


QTRegion::QTRegion()
{
}


QTRegion::~QTRegion()
{
}


void QTRegion::AddObject (GraphicObject* obj)
{
   // if it is a valid object, then insert it into list
  if (obj != 0)
  {
	  objects.push_back(obj);
  }
 
}

void QTRegion::RemoveObject (GraphicObject* obj)
{  
	std::list<GraphicObject*>::iterator i = find(objects.begin(), objects.end(), obj);

	if (i != objects.end())
	{
		objects.erase(i);
	}

}